/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * ------------------------------------------------------------------------- */

#include <malloc.h>
#include <string.h>

#include "ioG722Encode.h"

static OperationParameters parameters;

static short oneFrame[160] =
{
    0x0000, 0x226A, 0xC054, 0x5364, 0xA562, 0x5442, 0xBEBA, 0x2482, 0xFDBA,
    0xDFB4, 0x3E08, 0xAD86, 0x5A8E, 0xAAEE, 0x42D6, 0xD96A, 0x048C, 0x1E28,
    0xC3A6, 0x5182, 0xA58E, 0x55D4, 0xBBA6, 0x28A2, 0xF92E, 0xE400, 0x3AA2,
    0xAF84, 0x5A46, 0xA978, 0x45D2, 0xD558, 0x0918, 0x19D4, 0xC71E, 0x4F6A,
    0xA5F6, 0x572E, 0xB8C0, 0x2CA6, 0xF4A6, 0xE85E, 0x3716, 0xB1B6, 0x59C2,
    0xA83A, 0x48A4, 0xD162, 0x0D9C, 0x156E, 0xCABC, 0x4D1E, 0xA696, 0x5850,
    0xB606, 0x308E, 0xF024, 0xECCA, 0x3368, 0xB41A, 0x5904, 0xA734, 0x4B44,
    0xCD8A, 0x1218, 0x10FA, 0xCE7C, 0x4AA2, 0xA770, 0x5938, 0xB37C, 0x3456,
    0xEBAE, 0xF144, 0x2F98, 0xB6B0, 0x580C, 0xA668, 0x4DB6, 0xC9D2, 0x1688,
    0x0C7C, 0xD25C, 0x47F4, 0xA884, 0x59E8, 0xB124, 0x37FE, 0xE744, 0xF5C6,
    0x2BA8, 0xB974, 0x56DC, 0xA5D6, 0x4FF4, 0xC63E, 0x1AEA, 0x07F6, 0xD65A,
    0x4518, 0xA9D0, 0x5A5E, 0xAF00, 0x3B80, 0xE2EA, 0xFA50, 0x279C, 0xBC68,
    0x5574, 0xA57E, 0x51FE, 0xC2CE, 0x1F3A, 0x036A, 0xDA74, 0x4210, 0xAB54,
    0x5A98, 0xAD10, 0x3EDC, 0xDEA4, 0xFEDE, 0x2378, 0xBF86, 0x53D4, 0xA560,
    0x53D4, 0xBF86, 0x2376, 0xFEDE, 0xDEA4, 0x3EDC, 0xAD10, 0x5A98, 0xAB54,
    0x420E, 0xDA74, 0x0368, 0x1F3A, 0xC2CE, 0x5200, 0xA57E, 0x5574, 0xBC68,
    0x279C, 0xFA50, 0xE2EA, 0x3B80, 0xAF00, 0x5A5E, 0xA9D0, 0x4518, 0xD65A,
    0x07F4, 0x1AEA, 0xC63C, 0x4FF4, 0xA5D6, 0x56DC, 0xB976
};

/* Define function prototypes for the features we will override from the base */
static void g722Read(pSourceSink source, unsigned char *buffer,
                     uint32_t maxsize);

static void g722Write(pSourceSink sink, unsigned char *buffer, uint32_t size);

static int offset = 0;

/* ----------------------------------------------------------------------------
 * Function     : makeG722EncodeSourceSink
 * ----------------------------------------------------------------------------
 * Description  : This IO provider allows test data to be input and output
 *                to the codec and is provided for debug purposes. This
 *                method allows the construction of a suitable control block
 *                using malloc.
 * Inputs       : None
 * Outputs      : Pointer to an IO Source/Sink control block which has been
 *                configured to match the test data indicated above.
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
pSourceSink makeG722EncodeSourceSink(void)
{
    pSourceSink object = (pSourceSink)malloc(sizeof(struct _SourceSink));
    if (object != NULL)
    {
        object->read = &g722Read;
        object->write = &g722Write;
    }

    parameters.action = ENCODE | ENCODE_RESET;
    parameters.frameSize = 80;
    parameters.blockSize = 4;
    parameters.modeAndChannel = 1;
    offset = 0;

    return object;
}

/* ----------------------------------------------------------------------------
 * Function     : g722Read
 * ----------------------------------------------------------------------------
 * Description  : Implementation of the read method to provide data for test
 * Inputs       : source : pointer to a source/sink control block
 *                buffer : the buffer which will be used to return the test
 *                data
 *                maxsize : The maximum size of the buffer which must not be
 *                exceeded
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void g722Read(pSourceSink source, unsigned char *buffer,
                     uint32_t maxsize)
{
    memcpy(buffer, &oneFrame[offset * 80], parameters.frameSize * 2);
    memcpy(&source->operation, &parameters, sizeof(parameters));
    source->srcUsed = parameters.frameSize * 2;
    parameters.action = ENCODE;
    offset = 1 - offset;
}

/* ----------------------------------------------------------------------------
 * Function     : g722Write
 * ----------------------------------------------------------------------------
 * Description  : Implementation of the write method to verify test results
 * Inputs       : source : pointer to a source/sink control block
 *                buffer : the buffer which has been used to provide the
 *                test results
 *                size : The used size of the buffer
 * Outputs      : None
 * Assumptions  : None
 * ------------------------------------------------------------------------- */
static void g722Write(pSourceSink sink, unsigned char *buffer, uint32_t size)
{
    /* don't need checks here, using debugger to verify correct behaviour */
}
