/*                     v3.0 - 10/Jan/2007
  ============================================================================

                          U    U   GGG    SSSS  TTTTT
                          U    U  G       S       T
                          U    U  G  GG   SSSS    T
                          U    U  G   G       S   T
                           UUU     GG     SSS     T

                   ========================================
                    ITU-T - USER'S GROUP ON SOFTWARE TOOLS
                   ========================================


       =============================================================
       COPYRIGHT NOTE: This source code, and all of its derivations,
       is subject to the "ITU-T General Public License". Please have
       it  read  in    the  distribution  disk,   or  in  the  ITU-T
       Recommendation G.191 on "SOFTWARE TOOLS FOR SPEECH AND  AUDIO
       CODING STANDARDS". 
       ** This code has  (C) Copyright by CNET Lannion A TSS/CMC **
       =============================================================


MODULE:         USER-LEVEL FUNCTIONS FOR THE UGST G.722 MODULE

ORIGINAL BY:
   Simao Ferraz de Campos Neto
   COMSAT Laboratories                    Tel:    +1-301-428-4516
   22300 Comsat Drive                     Fax:    +1-301-428-9287
   Clarksburg MD 20871 - USA              E-mail: simao.campos@labs.comsat.com
    
   History:
History:
~~~~~~~~
14.Mar.95  v1.0       Released for use ITU-T UGST software package Tool
                      based on the CNET's 07/01/90 version 2.00
01.Jul.95  v2.0       Changed function declarations to work with many compilers;
                      reformated <simao@ctd.comsat.com>
10.Jan.07  v3.0       Updated with STL2005 v2.2 basic operators
                      <{balazs.kovesi,stephane.ragot}@orange-ftgroup.com>
Porting on LPDSP32 - Prithwee Narayana Bhat  
24.Nov.15 - LPDSP32 optimizations by Ivo Coenen:
- (32-bit) int for Word16 type instead of (16-bit) short
- mult_long multiplications in qmf_tx and qmf_rx functions
- indexing instead of pointer incrementing inside loops to get dualld instructions
- using circular buffers for FIR delay lines
- static inline functions
- hardware saturation in qmf_tx and qmf_rx functions
- shifting inside limit function
- binary search in quantl function
- numerous minor optimizations
- input/output code is unsigned char type for easier handling
  ============================================================================
*/
#include "operg722.h"
#include "g722.h"
#include "g722_com.h"
// #include "stl.h"
#include "funcg722.h"

void g722_reset_encoder (g722_state *encoder)
{
  Word16          xl, il;
  Word16          xh, ih;
  int             j;

  xl = xh = 0;

  for(j = 0; j < 24; j++) chess_loop_count(24)
  {
    encoder->qmf_tx_delayx[j] = 0;
  }
  encoder->p_tx_delayx = encoder->qmf_tx_delayx;

  il = lsbcodecrs (encoder);
  ih = hsbcodecrs (encoder);

}
/* .................... end of g722_reset_encoder() ....................... */


Word32 g722_encode (short *incode, unsigned char *code, int read1, 
		       g722_state *encoder, int bps)
{
  /* Encoder variables */
  Word16          xl, il;
  Word16          xh, ih;
  Word16          xin0, xin1;
  Word16          shift = 8 - bps;

  /* Auxiliary variables */
  int             i;

  /* Divide sample counter by 2 to account for QMF operation */
  read1 = read1 >> 1;

  /* Main loop - never reset */

  for(i = 0; i < read1; i++) chess_loop_range(2,MAX_FR_SIZE)
  {
    xin1 = *incode++;
    xin0 = *incode++;

    /* Calculation of the synthesis QMF samples */
    qmf_tx (xin0, xin1, &xl, &xh, encoder);

    /* Call the upper and lower band ADPCM encoders */
    il = lsbcod (xl, encoder);
    ih = hsbcod (xh, encoder);

    /* Mount the output G722 codeword: bits 0 to 5 are the lower-band
     * portion of the encoding, and bits 6 and 7 are the upper-band
     * portion of the encoding */
    code[i] = (unsigned char)(ih<<6 | il) >> shift;
  }

  /* Return number of samples read divided by 2 (=number of codes generated) */
  return(read1);
}
/* .................... end of g722_encode() .......................... */


void g722_reset_decoder (g722_state *decoder)
{
  Word16          il, ih;
  Word16          rl, rh;
  int             j;

  il = ih = 0;

  for(j = 0; j < 24; j++) chess_loop_count(24)
  {
    decoder->qmf_rx_delayx[j] = 0;
  }
  decoder->p_rx_delayx = decoder->qmf_rx_delayx;

  rl = lsbcodecrs (decoder);
  rh = hsbcodecrs (decoder);
}
/* .................... end of g722_reset_decoder() ....................... */


short g722_decode (unsigned char *code, short *outcode, int mode, 
		       int read1, g722_state *decoder)
{
  /* Decoder variables */
  Word16          il, ih;
  Word16          rl, rh;
  Word16          xout1, xout2;
  Word16          shifted;
  Word16          shift = mode - 1;
    
  /* Auxiliary variables */
  int             i;  

  /* Decode - reset is never applied here */
  for(i = 0; i < read1; i++)
  {
    /* Separate the input G722 codeword: bits 0 to 5 are the lower-band
     * portion of the encoding, and bits 6 and 7 are the upper-band
     * portion of the encoding */
    shifted = (short)(code[i] << shift);
    il = shifted & 0x3F; /* 6 bits of low SB */
    ih = (shifted >> 6) & 3;   /* 2 bits of high SB */

    /* Call the upper and lower band ADPCM decoders */
    rl = lsbdec (il, mode, decoder);
    rh = hsbdec (ih, decoder);

    /* Calculation of output samples from QMF filter */
    qmf_rx (rl, rh, &xout1, &xout2, decoder);

    /* Save samples in output vector */
    *outcode++ = xout1;
    *outcode++ = xout2;
  }

  /* Return number of samples generated */
  return(read1 << 1);
}
/* .................... end of g722_decode() .......................... */


