/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * -----------------------------------------------------------------------------
 */

#include "..\wrapper\wrapper.h"
#include "codecs/baseDSP/baseDSPCodecInternal.h"

#include "g722.h"

#define DEFAULT_FRAME_SIZE        (160)

/* we allow for a maximum of two channels to enable stereo (dual mono)
 * processing
 */
#define MAXIMUM_NUMBER_CHANNELS    (2)

typedef struct
{
    uint8_t frameSize;
    uint8_t blockSize;
    uint8_t mode;
    uint8_t channel;
    uint8_t packed;
    uint8_t bitsPerSample;
    uint8_t usedBits;
    uint16_t buffer;
    g722_state state;
} g722_config;

typedef struct
{
    g722_config encoder;
    g722_config decoder;
} g722EncodeDecode;

/* Statically create a couple of encoder states (for stereo operation) */
static g722EncodeDecode codecs[MAXIMUM_NUMBER_CHANNELS];

/* statically define a buffer to be used when packing and unpacking */
static unsigned char tempBuffer[DEFAULT_FRAME_SIZE];

/* ----------------------------------------------------------------------------
 * Function : reset
 * ----------------------------------------------------------------------------
 * Description : static method to reset a g722_config object to a known state
 *
 * Inputs :
 *         config : pointer to the G722 configuration object to be reset
 *         pConfig : pointer to a codec configuration object which specifies the
 *         operational parameters to be used.
 *
 * Outputs :
 *         config: the contents of the G722 configuration object will be updated
 *         in accordance with the specified operational parameters
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static void reset(g722_config *config, pBaseDSPCodec pConfig)
{
    volatile OperationParameters *op = &pConfig->parent.operation;
    config->frameSize = op->frameSize;
    config->blockSize = op->blockSize;
    config->mode = op->modeAndChannel & 0x0F;
    config->packed = (op->action & (PACKED | ALT_PACKED));
    config->bitsPerSample = 9 - config->mode;
    config->usedBits = 0;
    config->buffer = 0;
}

/* ----------------------------------------------------------------------------
 * Function : configure
 * ----------------------------------------------------------------------------
 * Description : Implements the configure method defined in "wrapper.h"
 * Inputs : refer "wrapper.h"
 * Outputs : refer "wrapper.h"
 * Assumptions : refer "wrapper.h"
 * ----------------------------------------------------------------------------
 */
void configure(pBaseDSPCodec /* unused */)
{
    /* Nothing required here */
}

/* ----------------------------------------------------------------------------
 * Function : getEncoderState
 * ----------------------------------------------------------------------------
 * Description : static method to get the current encoder state depending on
 * the selected channel
 *
 * Inputs :
 *         pConfig : pointer to a codec configuration object which specifies the
 *         operational parameters to be used.
 *
 * Outputs :
 *         returns a pointer to the required G722 encode configuration object
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static g722_config *getEncoderState(pBaseDSPCodec pConfig)
{
    volatile OperationParameters *op = &pConfig->parent.operation;
    int channel = (op->modeAndChannel >> 4) & 0x0F;
    return &codecs[channel].encoder;
}

/* ----------------------------------------------------------------------------
 * Function : encodeReset
 * ----------------------------------------------------------------------------
 * Description : Implements the encodeReset method defined in "wrapper.h"
 * Inputs : refer "wrapper.h"
 * Outputs : refer "wrapper.h"
 * Assumptions : refer "wrapper.h"
 * ----------------------------------------------------------------------------
 */
void encodeReset(pBaseDSPCodec pConfig)
{
    g722_config *myState = getEncoderState(pConfig);
    reset(myState, pConfig);
    g722_reset_encoder(&myState->state);
}

/* ----------------------------------------------------------------------------
 * Function : getDecoderState
 * ----------------------------------------------------------------------------
 * Description : static method to get the current decoder state depending on
 * the selected channel
 *
 * Inputs :
 *         pConfig : pointer to a codec configuration object which specifies the
 *         operational parameters to be used.
 *
 * Outputs :
 *         returns a pointer to the required G722 decode configuration object
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static g722_config *getDecoderState(pBaseDSPCodec pConfig)
{
    volatile OperationParameters *op = &pConfig->parent.operation;
    int channel = (op->modeAndChannel >> 4) & 0x0F;
    return &codecs[channel].decoder;
}

/* ----------------------------------------------------------------------------
 * Function : decodeReset
 * ----------------------------------------------------------------------------
 * Description : Implements the decodeReset method defined in "wrapper.h"
 * Inputs : refer "wrapper.h"
 * Outputs : refer "wrapper.h"
 * Assumptions : refer "wrapper.h"
 * ----------------------------------------------------------------------------
 */
void decodeReset(pBaseDSPCodec pConfig)
{
    g722_config *myState = getDecoderState(pConfig);
    reset(myState, pConfig);
    g722_reset_decoder(&myState->state);
}

/* ----------------------------------------------------------------------------
 * Function : encodeBuffer
 * ----------------------------------------------------------------------------
 * Description : Encodes a buffer of samples based on the G722 configuration
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the encode
 *
 *         input : pointer to a buffer of samples to be encoded
 *
 *         output : pointer to a buffer in which to write the encoded samples
 *
 * Outputs :
 *         output : the encoded samples will be written to the output buffer
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static void encodeBuffer(g722_config *context, short *input,
        unsigned char *output)
{
    int fSize = context->frameSize;
    int bSize = context->blockSize;
    int bps = context->bitsPerSample;

    for (int blockIn = 0; blockIn < fSize; blockIn += bSize)
        g722_encode(&input[blockIn], &output[blockIn >> 1], bSize,
                &context->state, bps);
}

/* ----------------------------------------------------------------------------
 * Function : packBuffer
 * ----------------------------------------------------------------------------
 * Description : packs a buffer of encoded samples
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the packing
 *
 *         input : pointer to a buffer of samples to be packed
 *
 *         output : pointer to a buffer in which to write the packed samples
 *
 * Outputs :
 *         output : the packed samples will be written to the output buffer
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static void packBuffer(g722_config *context, unsigned char *input,
        unsigned char *output)
{
    unsigned char *src = input;
    unsigned char *dst = output;

    uint16_t buffer = context->buffer;
    unsigned char used = context->usedBits;
    unsigned char bps = context->bitsPerSample;

    for (int i = 0; i < context->frameSize / 2; i++)
    {
        buffer |= (uint16_t) *src++ << used;
        used += bps;
        if (used >= 8)
        {
            *dst++ = (unsigned char) buffer;
            used -= 8;
            buffer >>= 8;
        }
    }

    context->buffer = buffer;
    context->usedBits = used;
}

/* ----------------------------------------------------------------------------
 * Function : packBufferAlt
 * ----------------------------------------------------------------------------
 * Description : packs a buffer of encoded samples specifically to match the
 *         current implementation of remote_mic_trx (Mode 3)
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the packing
 *
 *         input : pointer to a buffer of samples to be packed
 *
 *         output : pointer to a buffer in which to write the packed samples
 *
 * Outputs :
 *         output : the packed samples will be written to the output buffer
 *
 * Assumptions :
 *         context->framesize must be a multiple of four bytes
 *         context->bitsPerSample must be six
 *         bits being packed are already shifted into lower six bits
 * ----------------------------------------------------------------------------
 */
static void packBufferAlt(g722_config *context, unsigned char *input,
        unsigned char *output)
{
    unsigned char *src = input;
    unsigned char *dst = output;

    for (int i = 0; i < context->frameSize / 2; i += 4)
    {
    	*dst = *src++ << 2;
    	*dst++ |= *src >> 4;

    	*dst = *src++ << 4;
    	*dst++ |= *src >> 2;

    	*dst = *src++ << 6;
    	*dst++ |= *src++;
    }
}

/* ----------------------------------------------------------------------------
 * Function : expectedEncodedBytes
 * ----------------------------------------------------------------------------
 * Description : static method to provide the expected number of encoded bytes
 * based on the frame size and bits per sample
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the encode
 *
 * Outputs :
 *        The expected number of encoded and packed bytes
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static int expectedEncodedBytes(g722_config *context)
{
    int expectedLength = context->frameSize / 2;
    if (context->packed != 0)
        if (context->bitsPerSample == 7)
            expectedLength -= (expectedLength / 8);
        else if (context->bitsPerSample == 6)
            expectedLength -= (expectedLength / 4);
    return expectedLength;
}

/* ----------------------------------------------------------------------------
 * Function : encode
 * ----------------------------------------------------------------------------
 * Description : Implements the encode method defined in "wrapper.h"
 * Inputs : refer "wrapper.h"
 * Outputs : refer "wrapper.h"
 * Assumptions : refer "wrapper.h"
 * ----------------------------------------------------------------------------
 */
void encode(pBaseDSPCodec pConfig, unsigned char *inBuffer,
        unsigned char *outBuffer)
{
    g722_config *myState = getEncoderState(pConfig);
    if ((myState->bitsPerSample < 8) && (myState->packed != 0))
    {
        encodeBuffer(myState, (short *) inBuffer, tempBuffer);
        if (myState->packed == PACKED)
        	packBuffer(myState, tempBuffer, outBuffer);
        else
        	packBufferAlt(myState, tempBuffer, outBuffer);
    }
    else
    {
        encodeBuffer(myState, (short *) inBuffer, outBuffer);
    }
    pConfig->parent.outputBufferUsedSize = expectedEncodedBytes(myState);
}

/* ----------------------------------------------------------------------------
 * Function : decodeBuffer
 * ----------------------------------------------------------------------------
 * Description : Decodes a buffer of samples based on the G722 configuration
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the decode
 *
 *         input : pointer to a buffer of samples to be decoded
 *
 *         output : pointer to a buffer in which to write the decoded samples
 *
 * Outputs :
 *         output : the decoded samples will be written to the output buffer
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static void decodeBuffer(g722_config *context, unsigned char *input,
        short *output)
{
    int fSize = context->frameSize / 2;
    int bSize = context->blockSize / 2;
    for (int blockIn = 0; blockIn < fSize; blockIn += bSize)
        g722_decode(&input[blockIn], &output[blockIn << 1], context->mode,
                bSize, &context->state);
}

/* ----------------------------------------------------------------------------
 * Function : unpackBuffer
 * ----------------------------------------------------------------------------
 * Description : unpacks a buffer of packed encoded samples
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the unpacking
 *
 *         input : pointer to a buffer of packed encoded samples to be unpacked
 *
 *         output : pointer to a buffer in which to write the unpacked samples
 *
 * Outputs :
 *         output : the unpacked samples will be written to the output buffer
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static void unpackBuffer(g722_config *context, unsigned char *input,
        unsigned char *output)
{
    unsigned char *src = input;
    unsigned char *dst = output;

    unsigned char bps = context->bitsPerSample;
    uint16_t buffer = context->buffer;
    unsigned char used = context->usedBits;

    for (int i = 0; i < context->frameSize / 2; i++)
    {
        if (used < bps)
        {
            buffer |= ((uint16_t) *src++) << used;
            used += 8;
        }
        *dst++ = buffer;
        buffer >>= bps;
        used -= bps;
    }

    context->buffer = buffer;
    context->usedBits = used;
}

/* ----------------------------------------------------------------------------
 * Function : unpackBufferAlt
 * ----------------------------------------------------------------------------
 * Description : unpacks a buffer of packed encoded samples, mirror of
 *         packBufferAlt
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the unpacking
 *
 *         input : pointer to a buffer of packed encoded samples to be unpacked
 *
 *         output : pointer to a buffer in which to write the unpacked samples
 *
 * Outputs :
 *         output : the unpacked samples will be written to the output buffer
 *
 * Assumptions :
 *         All the same assumptions as
 * ----------------------------------------------------------------------------
 */
static void unpackBufferAlt(g722_config *context, unsigned char *input,
        unsigned char *output)
{
    unsigned char *src = input;
    unsigned char *dst = output;

    for (int i = 0; i < context->frameSize / 2; i += 3)
    {
        *dst++ = *src >> 2;

        *dst = (*src++ & 0x03) << 4;
        *dst++ |= (*src & 0xF0) >> 4;

        *dst = (*src++ & 0x0F) << 2;
        *dst++ |= (*src & 0xC0) >> 6;

        *dst++ = *src++ & 0x3F;
    }
}

/* ----------------------------------------------------------------------------
 * Function : expectedDecodedBytes
 * ----------------------------------------------------------------------------
 * Description : static method to provide the expected number of decoded bytes
 * based on the frame size
 *
 * Inputs :
 *         context : pointer to a G722 configuration object which provides the
 *         state for the decode
 *
 * Outputs :
 *        The expected number of decoded bytes
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
static int expectedDecodedBytes(g722_config *context)
{
    return context->frameSize * 2;
}

/* ----------------------------------------------------------------------------
 * Function : decode
 * ----------------------------------------------------------------------------
 * Description : Implements the decode method defined in "wrapper.h"
 * Inputs : refer "wrapper.h"
 * Outputs : refer "wrapper.h"
 * Assumptions : refer "wrapper.h"
 * ----------------------------------------------------------------------------
 */
void decode(pBaseDSPCodec pConfig, unsigned char *inBuffer,
        unsigned char *outBuffer, bool /* unused */)
{
    g722_config *myState = getDecoderState(pConfig);
    if ((myState->bitsPerSample < 8) && (myState->packed != 0))
    {
    	if (myState->packed == PACKED)
    		unpackBuffer(myState, inBuffer, tempBuffer);
    	else
    		unpackBufferAlt(myState, inBuffer, tempBuffer);
        decodeBuffer(myState, tempBuffer, (short *) outBuffer);
    }
    else
    {
        decodeBuffer(myState, inBuffer, (short *) outBuffer);
    }
    pConfig->parent.outputBufferUsedSize = expectedDecodedBytes(myState);
}

