/* ----------------------------------------------------------------------------
 * Copyright (c) 2017 Semiconductor Components Industries, LLC (d/b/a
 * ON Semiconductor), All Rights Reserved
 *
 * This code is the property of ON Semiconductor and may not be redistributed
 * in any form without prior written permission from ON Semiconductor.
 * The terms of use and warranty for this code are covered by contractual
 * agreements between ON Semiconductor and the licensee.
 *
 * This is Reusable Code.
 *
 * -----------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdint.h>

// Define the location of the CSS command register, this would normally be
// taken from a standard include file however to provide clarity in this
// example we will define it explicitly
#define CSS_CMD                         0xC00004

// Define the specific interrupt that the ARM will trigger when it wants
// us to perform a calculation
#define CSS_CMD_0                       (1<<0)

// Define the shared memory between the ARM and LPDSP processors
uint32_t chess_storage(DMB:0x00803800) sharedMemory[2];

// Define the interrupt register to notify the ARM of a completed operation
volatile static unsigned char chess_storage(DMIO:CSS_CMD) CssCmdGen;

// handling requests from the CM3 is activated
static volatile int actionRequired;

/* ----------------------------------------------------------------------------
 * Function : isr0
 * ----------------------------------------------------------------------------
 * Description : Interrupt service routine invoked on request from the CM3
 * Inputs : None
 * Outputs : sets flag indicating work needs to be done
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
extern "C" void isr0() property (isr) {
	// raise the flag indicating something needs to be processed
	actionRequired = true;
}

/* ----------------------------------------------------------------------------
 * Function : fibonacci
 * ----------------------------------------------------------------------------
 * Description : calculate the nth fibonacci number where:
 *                  0 is 0,
 *                  1 is 1,
 *                  2 is 1,
 *                  3 is 2, etc
 *
 * Inputs : nth, the number of the fibonacci sequence to return
 *
 * Outputs : the fibonacci number as defined above, or 0xFFFFFFFF if overflow
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
uint32_t fibonacci(uint32_t nth) {
    
    // we know the 48th fibonacci number will be too large for a uint32_t
    if (nth >= 48)
        return 0xFFFFFFFF;

    // deal with trivial cases
    if ((nth == 0) || (nth == 1))
        return nth;

    // otherwise calculate the return value in a loop
    uint32_t value = 0, first = 0, second = 1;
    while (nth-- > 1) {
        value = first + second;
        first = second;
        second = value;
    }
    return value;
}

/* ----------------------------------------------------------------------------
 * Function : main
 * ----------------------------------------------------------------------------
 * Description : The main entry point for the program
 *
 * 	Note: this halts the core when not busy and is awoken by an interrupt
 *      from the CM3
 *
 * Inputs : None
 *
 * Outputs : returns zero but the code is not expected to ever complete
 *
 * Assumptions : None
 * ----------------------------------------------------------------------------
 */
extern "C" int main(void) {
    
    // assume nothing ready to be processed yet
    actionRequired = false;

    // enable the interrupts
    enable_interrupts();
    
    // spin forever, waiting for interrupts from the CM3
    while (true) {        
        core_halt();

        // only do something if we are responding to an interrupt from the CM3
        if (actionRequired == true) {
            actionRequired = false;

            // calculate the nth fibonacci number
            sharedMemory[1] = fibonacci(sharedMemory[0]);
            
            // notify the CM3 that we have completed processing
            CssCmdGen = CSS_CMD_0;
        }

    }
    
    return 0;
}
